----------------------------------------------------------------------
-- Finite Field Chapter 6 package body (finite_field.adb)
-- 
-- Defines useful funtions for finite field operations
-- of Chapter 6 examples
---------------------------------------------------------------------
package body Finite_Fields is

   function B_Table(M: Digit) return Digit_Vector is
      B: Digit_Vector;
   begin
      B(0) := 1; 
      for I in 1 .. S-1 loop B(I) := (2**(I*K)) mod M; end loop;
      return B;
   end B_Table;

   function Mod_M_Addition(X, Y, M, K: Natural) return Natural is
      Z1, Z2, C1, C2: Natural;
   begin
      Z1 := X + Y; Z2 := (Z1 mod 2**K) + (2**K - M);
      C1 := Z1/2**K; C2 := Z2/2**K; 
      if C1 = 0 and C2 = 0 then return (Z1 mod 2**K); else return (Z2 mod 2**K); end if;
   end Mod_M_Addition;

   procedure Csa_Mod_Addition(Xc, Xs, Y: in Natural; Rc, Rs: out Natural) is
      procedure Csa(A, B, C: in Natural; Next_A, Next_B: out Natural) is
         Binary_A, Binary_B, Binary_C, Binary_Next_A, Binary_Next_B: Bit_Vector(0 .. K+1);
         Quotient, Accumulator: Natural;
      begin
         --natural to binary:
         Quotient := A; 
         for I in 0 .. K+1 loop Binary_A(I) := Quotient mod 2; Quotient := Quotient/2; end loop;
         Quotient := B;
         for I in 0 .. K+1 loop Binary_B(I) := Quotient mod 2; Quotient := Quotient/2; end loop;
         Quotient := C;
         for I in 0 .. K+1 loop Binary_C(I) := Quotient mod 2; Quotient := Quotient/2; end loop;
         --carry save addition:
         for I in 0 .. K+1 loop Binary_Next_A(I) := (Binary_A(I) + Binary_B(I) + Binary_C(I)) mod 2; end loop;
         Binary_Next_B(0) := 0;
         for I in 1 .. K+1 loop Binary_Next_B(I) := (Binary_A(I-1) + Binary_B(I-1) + Binary_C(I-1)) / 2; end loop;
         --binary to natural
         Accumulator := 0;
         for I in 0 .. K+1 loop Accumulator := Accumulator + Binary_Next_A(I)*(2**I); end loop;
         Next_A := Accumulator; 
         Accumulator := 0;
         for I in 0 .. K+1 loop Accumulator := Accumulator + Binary_Next_B(I)*(2**I); end loop;
         Next_B := Accumulator; 
      end Csa;
      function Quotient(Ss, Sc: in Natural) return Integer is
         Ss_High, Sc_High, T: Natural;
      begin
         Ss_High := Ss / (2**(K-1));   
         Sc_High := Sc / (2**(K-1)); 
         --with 4 bits:
         T := (Ss_High + Sc_High) mod 16;
         if T <= 3 then return 1;
         elsif T < 15 then return -1; 
         else return 0;
         end if;
      end Quotient;
      Ws, Wc, Minus_M: Natural;
   begin
      Csa(Xc, Xs, Y, Wc, Ws);
      Minus_M := (2**(K+3)) - M;
      if Quotient(Ws, Wc) = 1 then Csa(Wc, Ws, Minus_M, Rc, Rs);
      elsif Quotient(Ws, Wc) = 0 then Csa(Wc, Ws, 0, Rc, Rs);
      else Csa(Wc, Ws, M, Rc, Rs);
      end if;
      Rc := Rc mod 2**(K+2); Rs := Rs mod 2**(K+2);
   end Csa_Mod_Addition;

   procedure Csa_Mod_Doubling(Xc, Xs: in Natural; Rc, Rs: out Natural) is
      procedure Csa(A, B, C: in Natural; Next_A, Next_B: out Natural) is
         Binary_A, Binary_B, Binary_C, Binary_Next_A, Binary_Next_B: Bit_Vector(0 .. K+1);
         Quotient, Accumulator: Natural;
      begin
         --natural to binary:
         Quotient := A; 
         for I in 0 .. K+1 loop Binary_A(I) := Quotient mod 2; Quotient := Quotient/2; end loop;
         Quotient := B;
         for I in 0 .. K+1 loop Binary_B(I) := Quotient mod 2; Quotient := Quotient/2; end loop;
         Quotient := C;
         for I in 0 .. K+1 loop Binary_C(I) := Quotient mod 2; Quotient := Quotient/2; end loop;
         --carry save addition:
         for I in 0 .. K+1 loop Binary_Next_A(I) := (Binary_A(I) + Binary_B(I) + Binary_C(I)) mod 2; end loop;
         Binary_Next_B(0) := 0;
         for I in 1 .. K+1 loop Binary_Next_B(I) := (Binary_A(I-1) + Binary_B(I-1) + Binary_C(I-1)) / 2; end loop;
         --binary to natural
         Accumulator := 0;
         for I in 0 .. K+1 loop Accumulator := Accumulator + Binary_Next_A(I)*(2**I); end loop;
         Next_A := Accumulator; 
         Accumulator := 0;
         for I in 0 .. K+1 loop Accumulator := Accumulator + Binary_Next_B(I)*(2**I); end loop;
         Next_B := Accumulator; 
      end Csa;
      function Quotient(Ss, Sc: in Natural) return Integer is
         Ss_High, Sc_High, T: Natural;
      begin
         Ss_High := Ss / (2**(K-1));   
         Sc_High := Sc / (2**(K-1)); 
         --with 4 bits:
         T := (Ss_High + Sc_High) mod 16;
         if T <= 3 then return 1;
         elsif T < 15 then return -1; 
         else return 0;
         end if;
      end Quotient;
      Ws, Wc, Minus_M: Integer;
   begin
      Ws := 2*Xs; Wc := 2*Xc;
      Minus_M := (2**(K+3)) - M;
      if Quotient(Ws, Wc) = 1 then Csa(Wc, Ws, Minus_M, Rc, Rs);
      elsif Quotient(Ws, Wc) = 0 then Csa(Wc, Ws, 0, Rc, Rs);
      else Csa(Wc, Ws, M, Rc, Rs);
      end if;
      Rc := Rc mod 2**(K+2); Rs := Rs mod 2**(K+2);
   end Csa_Mod_Doubling;

   function Mp(X, Y: Natural) return Natural is
      Product, Q: Natural;
   begin
      Product := X*Y;
      Q := (Product + ((Product*Minus_Inv_M) mod R)*M) / R;
      if Q >= M then return Q-M; else return Q; end if; 
   end Mp;

   function Divide_By_2(X, P: in Integer) return Integer is
      Y: Integer;
   begin
      if (X mod 2) = 0 then Y := X/2; else Y := (X+P)/2; end if;
      return Y;
   end Divide_By_2;

   function Divide_By_4(X, P: in Integer) return Integer is
      Y: Integer;
   begin
      if (P mod 4) = 1 then
         if (X mod 4) = 0 then Y := X/4;
         elsif (X mod 4) = 1 then Y := (X - P)/4;
         elsif (X mod 4) = 2 then Y := (X + 2*P)/4;
         else Y := (X + P)/4;
         end if;
      elsif (P mod 4) = 3 then
         if (X mod 4) = 0 then Y := X/4;
         elsif (X mod 4) = 1 then Y := (X + P)/4;
         elsif (X mod 4) = 2 then Y := (X + 2*P)/4;
         else Y := (X - P)/4;
         end if;
      end if;
      return Y;
   end Divide_By_4;   

   function Invert(Y: Natural_Mod_P) return Natural_Mod_P is
      A, B, C, D, Q, R, Next_D: Integer;
   begin
      A := P; B := Y; C := 0; D := 1;
      while B > 1 loop
         Q := A/B; R := A mod B; Next_D := C - D*Q;
         A := B; B := R; C := D; D := Next_D;
      end loop;
      return (D mod P);
   end Invert;

   
end Finite_Fields;

