----------------------------------------------------------------------
-- Polynomials Chapter 6 package (polynomials.ads)
-- 
-- Defines useful funtions for polynomial operations
-- of Chapter 6 examples
---------------------------------------------------------------------

--with finite_fields; use finite_fields;
package polynomials is

   m: constant natural := 17;
   p: constant natural := 239;
   
   --m: constant natural := 4;
   --p: constant natural := 239;
   
   --m: constant natural := 97;
   --p: constant natural := 3;

   subtype Natural_Mod_P is Natural range 0 .. P-1;
   type Polynomial is array(0 .. M) of Natural_Mod_P;

   function Degree(A: Polynomial) return Natural;
   function Subtract(A, B: Polynomial) return Polynomial;
   function Invert(Y: Natural_Mod_P) return Natural_Mod_P;
   function Product(A: Polynomial; B: Natural_Mod_P) return Polynomial;
   function Shift_One(A: Polynomial) return Polynomial;
   function Divide_By_X(A, F: Polynomial) return Polynomial;
   function Add(A, B: Polynomial) return Polynomial;
   function Multiply_By_X(A, F: Polynomial) return Polynomial;
   function Multiply_By_X(A: Polynomial) return Polynomial;
   function Product_Mod_F(A, B, F: Polynomial) return Polynomial;
   function Quotient(Num, Den: Polynomial) return Polynomial;
   function Remainder(Num, Den: Polynomial) return Polynomial;
   function Shift(A: Polynomial; T: Natural) return Polynomial;
   function Shift(A, F: Polynomial; T: Natural) return Polynomial;
   function Pseudo_Quotient(Num, Den: Polynomial) return Polynomial;
   function Pseudo_Remainder(Num, Den: Polynomial) return Polynomial;
   procedure Swap(A, B: in out Polynomial); 
   procedure Swap(S, T: in out Natural); 
   --procedure pseudo_remainder(n_a: in polynomial; deg_a: in natural; 
   --   n_b: in polynomial; deg_b: in natural; n_r: out polynomial; deg_r: out natural);
   --procedure pseudo_remainder(m_a: in polynomial; deg_a: in natural; coef_a: in natural_mod_p;
   --   m_b: in polynomial; deg_b: in natural; coef_b: in natural_mod_p;
   --   m_r: out polynomial; deg_r: out natural; coef_r: out natural_mod_p); 
   function Frobenius(J, I: Natural)return Natural_Mod_P;
  
end polynomials;

