----------------------------------------------------------------------------
-- LSE first Mod f multiplier (LSE_first_Mod_f_Multiplier.vhd)
--
-- The hardware is genenerate for a specific P = 239.
-- And for an especified F
-- use Least Significant Element (LSE) First algorithm
--
-- for testing binary_algorithm_polynomials.vhd
----------------------------------------------------------------------------

------------------------------------------------------------
-- mod 239 reducer
------------------------------------------------------------
library ieee; use ieee.std_logic_1164.all;
use ieee.std_logic_arith.all;
use ieee.std_logic_unsigned.all;
entity mod_239_reducer is
port (
  x: in std_logic_vector(15 downto 0);
  z: out std_logic_vector(7 downto 0)
  );
end mod_239_reducer;

architecture circuit of mod_239_reducer is
  signal x1_by_17, x0, sum, xx1_by_17, xx0, xxx: std_logic_vector(8 downto 0);
  signal long_xxx, minus_239, dif: std_logic_vector(9 downto 0);
  signal x2_by_33, long_sum, xx: std_logic_vector(9 downto 0);
begin
  x1_by_17 <= '0'&x(11 downto 8)&x(11 downto 8);
  x0 <= '0'&x(7 downto 0);
  sum <= x1_by_17 + x0;
  x2_by_33 <= '0'&x(15 downto 12)&'0'&x(15 downto 12);
  long_sum <= '0'&sum;
  xx <= x2_by_33 + long_sum;
  xx1_by_17 <= "000"&xx(9 downto 8) &"00"&xx(9 downto 8);
  xx0 <= '0'&xx(7 downto 0);
  xxx <= xx1_by_17 + xx0;
  minus_239 <= conv_std_logic_vector(273, 10);
  long_xxx <= '0'&xxx;
  dif <= long_xxx + minus_239;
  with dif(9) select z <= dif(7 downto 0) when '1', xxx(7 downto 0) when others;
end circuit;

------------------------------------------------------------
-- subtractor mod P 
------------------------------------------------------------
library ieee; 
use ieee.std_logic_1164.all;
use ieee.std_logic_arith.all;
use ieee.std_logic_unsigned.all;
use work.binary_algorithm_polynomials_parameters.all;


entity subtractor_mod_P is
port (
  x, y: in std_logic_vector(K-1 downto 0);
  z: out std_logic_vector(K-1 downto 0)
  );
end subtractor_mod_P;

architecture rtl of subtractor_mod_P is
  signal z1, z2: std_logic_vector(K downto 0);
begin
  z1 <= ('0' & x) - y;
  z2 <= z1 + P;
  with z1(K) select 
       z <= z1(K-1 downto 0) when '0', 
            z2(K-1 downto 0) when others;
end rtl;

------------------------------------------------------------
-- mod f multiplier 
------------------------------------------------------------
library IEEE;
use IEEE.std_logic_1164.all;
use IEEE.std_logic_arith.all;
use IEEE.std_logic_unsigned.all;
use work.binary_algorithm_polynomials_parameters.all;

entity LSE_first_mod_f_mult_test2 is
port(
  a, b: in polynomial;
  clk, reset, start: in std_logic;  
  z: out polynomial;
  done: out std_logic
  );
end LSE_first_mod_f_mult_test2;

architecture circuit of LSE_first_mod_f_mult_test2 is
  component subtractor_mod_P is
  port (
    x, y: in std_logic_vector(K-1 downto 0);
    z: out std_logic_vector(K-1 downto 0) );
  end component;
  component mod_239_reducer is
  port(
    x: in std_logic_vector(15 downto 0);
    z: out std_logic_vector(7 downto 0) );
  end component;

  constant zero_coef: std_logic_vector(K-1 downto 0) := (others => '0');
  signal int_a, int_b, next_a, next_c, c, mult_sub: polynomial;
  type polynomial_2K is array(M-1 downto 0) of std_logic_vector(2*K-1 downto 0);
  signal mult_add, mult_f_x_a: polynomial_2K;
  constant ZERO: std_logic_vector(K-1 downto 0) := (others => '0'); 

  signal load, update, count_equal_zero: std_logic;
  signal count: std_logic_vector(logM-1 downto 0);
  type states is range 0 to 3;
  signal current_state: states;

begin

  next_c_calc: for i in 0 to m-1 generate
    mult_add(i) <= ( int_b(0) * int_a(i) ) + c(i);
    comp1: mod_239_reducer port map(mult_add(i), next_c(i));
  end generate;
  
  next_a_calc: for i in 1 to m-1 generate
    mult_f_x_a(i) <= F(i) * int_a(M-1);
    comp1: mod_239_reducer port map (mult_f_x_a(i), mult_sub(i));
    comp2: subtractor_mod_P port map (int_a(i-1), mult_sub(i), next_a(i));
  end generate;
  mult_f_x_a(0) <= ( F(0) * int_a(M-1) );
  comp1: mod_239_reducer port map(mult_f_x_a(0), mult_sub(0));
  comp2: subtractor_mod_P port map(ZERO, mult_sub(0), next_a(0));

  registers_abc: process(clk)
  begin
  if clk'event and clk = '1' then
    if load = '1' then 
      c <= ZERO_POLY;
      int_b <= b;
      int_a <= a;
    elsif update = '1' then 
      c <= next_c;
      int_b <= zero_coef & int_b(M-1 downto 1);
      int_a <= next_a;
    end if;
  end if;
  end process registers_abc;

  z <= c;

  counter: process(clk)
  begin
  if clk'event and clk = '1' then
    if load = '1' then count <= conv_std_logic_vector(M-1, LOGM);
    elsif update = '1' then count <= count - 1;
    end if;
  end if;
  end process counter;

  count_equal_zero <= '1' when count = 0 else '0';

  control_unit: process(clk, reset, count_equal_zero, current_state)
  begin
  case current_state is
    when 0 to 1 => load <= '0'; update <= '0'; done <= '1';
    when 2 => load <= '1'; update <= '0'; done <= '0';
    when 3 => load <= '0'; update <= '1'; done <= '0';
  end case;

  if reset = '1' then current_state <= 0;
  elsif clk'event and clk = '1' then
    case current_state is
      when 0 => if start = '0' then current_state <= 1; end if;
      when 1 => if start = '1' then current_state <= 2; end if;
      when 2 => current_state <= 3;
      when 3 => if count_equal_zero = '1' then current_state <= 0; end if;
    end case;
  end if;

  end process control_unit;

end circuit;
