------------------------------------------------------------
-- adder-subtractor
----------------------------------------------------------------------------------------------------------------------------------------
-- Optimal Extension Field divider (oef.vhd)
--
-- Defines the following entities:
--    - reduction_to_multiplications_parameters (a package)
--    - mod_239_inverter (read the table and gives the mod 239 inversion)
--    - mod_239_reducer (as in ch3) 
--    - mod_239_multiplier (as in ch3) 

--    - subtractor parameters (a package for subtractor) 
--    - subtractor (defines a modular subtraction) 
--    - pseudo_Euclidean_divider_parameters (a package)
--    - pseudo_Euclidean_divider (the divider)
--
----------------------------------------------------------------------------
 
------------------------------------------------------------
-- OEF_parameters
------------------------------------------------------------
library IEEE;
use IEEE.std_logic_1164.all;
use IEEE.std_logic_arith.all;
use IEEE.std_logic_unsigned.all;
package oef_parameters is
  constant k: natural := 8;
  constant p: std_logic_vector(k-1 downto 0) := conv_std_logic_vector(239, k);
  --subtype coefficient is natural range 0 to 238;
  constant m: natural := 17;
  --subtype degree is natural range 0 to m;
  type long_polynomial is array(m downto 0) of std_logic_vector(k-1 downto 0);
  type polynomial is array(m-1 downto 0) of std_logic_vector(k-1 downto 0);
  constant logm: natural := 6;
  constant f: long_polynomial :=
  ("00000001", "00000000", "00000000", "00000000", "00000000", "00000000", "00000000", "00000000", "00000000",
  "00000000", "00000000", "00000000", "00000000", "00000000", "00000000", "00000000", "00000000", "11101101");
  constant f1: polynomial :=
  (conv_std_logic_vector(67,8),conv_std_logic_vector(187,8),conv_std_logic_vector(101,8),conv_std_logic_vector(75,8),
  conv_std_logic_vector(6,8),conv_std_logic_vector(163,8),conv_std_logic_vector(166,8),conv_std_logic_vector(128,8),
  conv_std_logic_vector(211,8),conv_std_logic_vector(36,8),conv_std_logic_vector(22,8),conv_std_logic_vector(40,8),
  conv_std_logic_vector(51,8),conv_std_logic_vector(71,8),conv_std_logic_vector(216,8),conv_std_logic_vector(132,8),
  "00000001");
  constant f2: polynomial :=
  (conv_std_logic_vector(187,8),conv_std_logic_vector(75,8),conv_std_logic_vector(163,8),conv_std_logic_vector(128,8),
  conv_std_logic_vector(36,8),conv_std_logic_vector(40,8),conv_std_logic_vector(71,8),conv_std_logic_vector(132,8),
  conv_std_logic_vector(67,8),conv_std_logic_vector(101,8),conv_std_logic_vector(6,8),conv_std_logic_vector(166,8),
  conv_std_logic_vector(211,8),conv_std_logic_vector(22,8),conv_std_logic_vector(51,8),conv_std_logic_vector(216,8),
  "00000001");
  constant f4: polynomial :=
  (conv_std_logic_vector(75,8),conv_std_logic_vector(128,8),conv_std_logic_vector(40,8),conv_std_logic_vector(132,8),
  conv_std_logic_vector(101,8),conv_std_logic_vector(166,8),conv_std_logic_vector(22,8),conv_std_logic_vector(216,8),
  conv_std_logic_vector(187,8),conv_std_logic_vector(163,8),conv_std_logic_vector(36,8),conv_std_logic_vector(71,8),
  conv_std_logic_vector(67,8),conv_std_logic_vector(6,8),conv_std_logic_vector(211,8),conv_std_logic_vector(51,8),
  "00000001");
  constant f8: polynomial :=
  (conv_std_logic_vector(128,8),conv_std_logic_vector(132,8),conv_std_logic_vector(166,8),conv_std_logic_vector(216,8),
  conv_std_logic_vector(163,8),conv_std_logic_vector(71,8),conv_std_logic_vector(6,8),conv_std_logic_vector(51,8),
  conv_std_logic_vector(75,8),conv_std_logic_vector(40,8),conv_std_logic_vector(101,8),conv_std_logic_vector(22,8),
  conv_std_logic_vector(187,8),conv_std_logic_vector(36,8),conv_std_logic_vector(67,8),conv_std_logic_vector(211,8),
  "00000001");
  --constant one_poly: polynomial := (
  --"00000000", "00000000", "00000000", "00000000", 
  --"00000000", "00000000", "00000000", "00000000", 
  --"00000000", "00000000", "00000000", "00000000", 
  --"00000000", "00000000", "00000000", "00000000", "00000001");
  constant zero_poly: polynomial := (others => "00000000");
  constant zero_coef: std_logic_vector(k-1 downto 0) := (others => '0');
end oef_parameters;


------------------------------------------------------------
-- mod 239 inverter
------------------------------------------------------------
library IEEE;
use IEEE.STD_LOGIC_1164.ALL;
use IEEE.STD_LOGIC_ARITH.ALL;
use IEEE.STD_LOGIC_UNSIGNED.ALL;

entity mod_239_inverter is
port (
  clk: in std_logic; 
  x: in std_logic_vector(7 downto 0);
  z: out std_logic_vector(7 downto 0)
);
end mod_239_inverter;

architecture table of mod_239_inverter is
  type RomType is array (0 to 238) of natural;
  constant ROM: RomType  := (0, 1, 120, 80, 60, 48, 40, 205, 30, 186, 24, 87, 20, 92, 222, 16,
    15, 225, 93, 151, 12, 148, 163, 52, 10, 153, 46, 62, 111, 33, 8, 54,
    127, 29, 232, 41, 166, 84, 195, 190, 6, 35, 74, 189, 201, 85, 26, 178,
    5, 200, 196, 75, 23, 230, 31, 113, 175, 130, 136, 158, 4, 192, 27, 129,
    183, 114, 134, 132, 116, 97, 140, 101, 83, 203, 42, 51, 217, 149, 95, 118,
    3, 180, 137, 72, 37, 45, 214, 11, 220, 94, 162, 218, 13, 18, 89, 78,
    122, 69, 100, 169, 98, 71, 157, 181, 131, 173, 115, 172, 135, 182, 176, 28,
    207, 55, 65, 106, 68, 143, 79, 237, 2, 160, 96, 171, 133, 174, 184, 32,
    211, 63, 57, 104, 67, 124, 66, 108, 58, 82, 168, 141, 70, 139, 170, 117,
    161, 150, 221, 226, 21, 77, 145, 19, 228, 25, 194, 202, 167, 102, 59, 236,
    121, 144, 90, 22, 188, 197, 36, 156, 138, 99, 142, 123, 107, 105, 125, 56,
    110, 212, 47, 235, 81, 103, 109, 64, 126, 208, 9, 216, 164, 43, 39, 234,
    61, 213, 154, 38, 50, 165, 204, 233, 49, 44, 155, 73, 198, 7, 210, 112,
    185, 231, 206, 128, 177, 193, 86, 229, 187, 76, 91, 227, 88, 146, 14, 224,
    223, 17, 147, 219, 152, 215, 53, 209, 34, 199, 191, 179, 159, 119, 238 );
begin
  z <= conv_std_logic_vector(ROM(conv_integer(x)),8);
end table;

------------------------------------------------------------
-- mod 239 reducer
------------------------------------------------------------
library ieee; use ieee.std_logic_1164.all;
use ieee.std_logic_arith.all;
use ieee.std_logic_unsigned.all;
entity mod_239_reducer is
port (
  x: in std_logic_vector(15 downto 0);
  z: out std_logic_vector(7 downto 0)
  );
end mod_239_reducer;

architecture circuit of mod_239_reducer is
  signal x1_by_17, x0, sum, xx1_by_17, xx0, xxx: std_logic_vector(8 downto 0);
  signal long_xxx, minus_239, dif: std_logic_vector(9 downto 0);
  signal x2_by_33, long_sum, xx: std_logic_vector(9 downto 0);
begin
  x1_by_17 <= '0'&x(11 downto 8)&x(11 downto 8);
  x0 <= '0'&x(7 downto 0);
  sum <= x1_by_17 + x0;
  x2_by_33 <= '0'&x(15 downto 12)&'0'&x(15 downto 12);
  long_sum <= '0'&sum;
  xx <= x2_by_33 + long_sum;
  xx1_by_17 <= "000"&xx(9 downto 8) &"00"&xx(9 downto 8);
  xx0 <= '0'&xx(7 downto 0);
  xxx <= xx1_by_17 + xx0;
  minus_239 <= conv_std_logic_vector(273, 10);
  long_xxx <= '0'&xxx;
  dif <= long_xxx + minus_239;
  with dif(9) select z <= dif(7 downto 0) when '1', xxx(7 downto 0) when others;
end circuit;

------------------------------------------------------------
-- mod 239 multiplier
------------------------------------------------------------
library ieee; use ieee.std_logic_1164.all;
use ieee.std_logic_arith.all;
use ieee.std_logic_unsigned.all;
entity mod_239_multiplier is
port (
  x, y: in std_logic_vector(7 downto 0);
  z: out std_logic_vector(7 downto 0)
);
end mod_239_multiplier;

architecture circuit of mod_239_multiplier is
  component mod_239_reducer is
  port (
    x: in std_logic_vector(15 downto 0);
    z: out std_logic_vector(7 downto 0) );
  end component;
  signal product: std_logic_vector(15 downto 0);
begin
  product <= x*y;
  main_component: mod_239_reducer port map(product, z);
end circuit;

------------------------------------------------------------
-- adder-subtractor
------------------------------------------------------------
library ieee; 
use ieee.std_logic_1164.all;
use ieee.std_logic_arith.all;
use ieee.std_logic_unsigned.all;
use work.oef_parameters.all;
entity adder_subtractor is
port (
  x, y: in std_logic_vector(k-1 downto 0);
  addb_sub: in std_logic;
  z: out std_logic_vector(k-1 downto 0)
  );
end adder_subtractor;

architecture rtl of adder_subtractor is
  signal long_x, xor_y, sum1, long_z1, xor_p, sum2: std_logic_vector(k downto 0);
  signal c1, c2, sel: std_logic;
  signal z1, z2: std_logic_vector(k-1 downto 0);
begin
  long_x <= '0'&x;
  xor_gates1: for i in 0 to k-1 generate xor_y(i) <= y(i) xor addb_sub; end generate;
  xor_y(k) <= '0';
  sum1 <= addb_sub + long_x + xor_y;
  c1 <= sum1(k);
  z1 <= sum1(k-1 downto 0);
  long_z1 <= '0'&z1;
  xor_gates2: for i in 0 to k-1 generate xor_p(i) <= p(i) xor not(addb_sub); end generate;
  xor_p(k) <= '0';
  sum2 <= not(addb_sub) + long_z1 + xor_p;
  c2 <= sum2(k);
  z2 <= sum2(k-1 downto 0);
  sel <= (not(addb_sub) and (c1 or c2)) or (addb_sub and not(c1));
  with sel select z <= z1 when '0', z2 when others;
end rtl;

------------------------------------------------------------
-- mod_f_multiplier
------------------------------------------------------------
library IEEE;
use IEEE.std_logic_1164.all;
use IEEE.std_logic_arith.all;
use IEEE.std_logic_unsigned.all;
use work.oef_parameters.all;
entity mod_f_multiplier is
port(
  a, b: in polynomial;
  clk, reset, start: in std_logic;  
  z: out polynomial;
  done: out std_logic
  );
end mod_f_multiplier;

architecture circuit of mod_f_multiplier is
component mod_239_multiplier is
port 
(x, y: in std_logic_vector(7 downto 0);
z: out std_logic_vector(7 downto 0)
);
end component;
component adder_subtractor is
port (
x, y: in std_logic_vector(k-1 downto 0);
addb_sub: in std_logic;
z: out std_logic_vector(k-1 downto 0)
);
end component;
signal c, add_in1, add_in2, mult_in1, mult_in2, next_c, int_b: polynomial;
signal serial_out: std_logic_vector(k-1 downto 0); 
signal sel, ce_c, load, update, count_equal_zero: std_logic;
signal count: std_logic_vector(logm-1 downto 0);
type states is range 0 to 4;
signal current_state: states;
begin
main_component: for i in 1 to m-1 generate
with sel select add_in1(i) <= c(i) when '0', c(i-1) when others;
with sel select mult_in1(i) <= c(m-1) when '1', serial_out when others;
with sel select mult_in2(i) <= f(i) when '1', a(i) when others;
component1: mod_239_multiplier port map(mult_in1(i), mult_in2(i), add_in2(i));
component2: adder_subtractor port map(add_in1(i), add_in2(i), sel, next_c(i));
end generate;
with sel select add_in1(0) <= c(0) when '0', zero_coef when others;
with sel select mult_in1(0) <= c(m-1) when '1', serial_out when others;
with sel select mult_in2(0) <= f(0) when '1', a(0) when others;
component1: mod_239_multiplier port map(mult_in1(0), mult_in2(0), add_in2(0));
component2: adder_subtractor port map(add_in1(0), add_in2(0), sel, next_c(0));
c_register: process(clk)
begin
if clk'event and clk = '1' then
if load = '1' then c <= zero_poly;
elsif ce_c = '1' then c <= next_c;
end if;
end if;
end process c_register;
z <= c;
shift_register: process(clk)
begin
if clk'event and clk = '1' then
if load = '1' then int_b <= b;
elsif update = '1' then
for i in m-1 downto 1 loop int_b(i) <= int_b(i-1); end loop;
int_b(0) <= zero_coef;
end if;
end if;
end process shift_register;
serial_out <= int_b(m-1);
a_counter: process(clk)
begin
if clk'event and clk = '1' then
if load = '1' then count <= conv_std_logic_vector(m-1, logm);
elsif update = '1' then count <= count - 1;
end if;
end if;
end process a_counter;
count_equal_zero <= '1' when count = conv_std_logic_vector(0, logm) else '0';
control_unit: process(clk, reset, count_equal_zero, current_state)
begin
case current_state is
when 0 to 1 => sel <= '0'; ce_c <= '0'; load <= '0'; update <= '0'; done <= '1';
when 2 => sel <= '0'; ce_c <= '0'; load <= '1'; update <= '0'; done <= '0';
when 3 => sel <= '1'; ce_c <= '1'; load <= '0'; update <= '0'; done <= '0';
when 4 => sel <= '0'; ce_c <= '1'; load <= '0'; update <= '1'; done <= '0';
end case;
if reset = '1' then current_state <= 0;
elsif clk'event and clk = '1' then
case current_state is
when 0 => if start = '0' then current_state <= 1; end if;
when 1 => if start = '1' then current_state <= 2; end if;
when 2 => current_state <= 3;
when 3 => current_state <= 4;
when 4 => if count_equal_zero = '1' then current_state <= 0; else current_state <= 3; end if;
end case;
end if;
end process control_unit;
end circuit;


------------------------------------------------------------
-- OEF
------------------------------------------------------------
library IEEE;
use IEEE.std_logic_1164.all;
use IEEE.std_logic_arith.all;
use IEEE.std_logic_unsigned.all;
use work.oef_parameters.all;
entity oef is
port(
  g, h: in polynomial;
  clk, reset, start: in std_logic;  
  z: out polynomial;
  done: out std_logic
  );
end oef;

architecture circuit of oef is
  component mod_239_inverter is
  port (
    clk: in std_logic; 
    x: in std_logic_vector(7 downto 0);
    z: out std_logic_vector(7 downto 0) );
  end component;
  component mod_239_multiplier is
  port (
    x, y: in std_logic_vector(7 downto 0);
    z: out std_logic_vector(7 downto 0) );
  end component;
  ------------------------------------
  --provisional
  ------------------------------------
  component mod_f_multiplier is
  port(
    a, b: in polynomial;
    clk, reset, start: in std_logic;  
    z: out polynomial;
    done: out std_logic );
  end component;
  ------------------------------------
  signal a, e, next_e, next_a, f_coef, op1, op2, ahg: polynomial;
  signal inv_a0, mult_out: std_logic_vector(k-1 downto 0);
  signal sel_f, sel_a: std_logic_vector(1 downto 0);
  signal sel_e, ce_a, ce_e, start_mult, mult_done, load: std_logic;

  type states is range 0 to 26;
  signal current_state: states;

begin

  main_iteration: for j in 1 to m-1 generate
    with sel_f select f_coef(j) <= f1(j) when "00", f2(j) when "01", f4(j) when "10", f8(j) when others;
    with sel_e select op1(j) <= a(j) when '0', e(j) when others;
    with sel_e select op2(j) <= f_coef(j) when '0', inv_a0 when others;
    mod_239_multipliers: mod_239_multiplier port map(op1(j), op2(j), next_e(j));
  end generate;
  mod_239_multiplier2: mod_239_multiplier port map(e(0), inv_a0, mult_out);
  with sel_e select next_e(0) <= a(0) when '0', mult_out when others;

  with sel_a select ahg <= a when "00", h when "01", g when others;

  mod_f_multiplier1: mod_f_multiplier port map(e, ahg, clk, reset, start_mult, next_a, mult_done);
  z <= next_a;

  inverter: mod_239_inverter port map(clk, a(0), inv_a0);

  register_e: process(clk)
  begin
  if clk'event and clk = '1' then
    if ce_e = '1' then e <= next_e; 
    end if;
  end if;
  end process;

  register_a: process(clk)
  begin
  if clk'event and clk = '1' then
    if load = '1' then a <= h;
    elsif ce_a = '1' then a <= next_a;
    end if;
  end if;
  end process;

  control_unit: process(clk, reset, current_state, mult_done)
  begin
  case current_state is
    when 0 to 1 => sel_f <= "00"; sel_e <= '0'; sel_a <= "00"; ce_e <= '0'; ce_a <= '0'; start_mult <= '0'; load <= '0'; done <= '1';
    when 2 => sel_f <= "00"; sel_e <= '0'; sel_a <= "00"; ce_e <= '0'; ce_a <= '0'; start_mult <= '0'; load <= '1'; done <= '0';

    when 3 => sel_f <= "00"; sel_e <= '0'; sel_a <= "00"; ce_e <= '1'; ce_a <= '0'; start_mult <= '0'; load <= '0'; done <= '0';
    when 4 => sel_f <= "00"; sel_e <= '0'; sel_a <= "00"; ce_e <= '0'; ce_a <= '0'; start_mult <= '1'; load <= '0'; done <= '0';
    when 5 => sel_f <= "00"; sel_e <= '0'; sel_a <= "00"; ce_e <= '0'; ce_a <= '0'; start_mult <= '0'; load <= '0'; done <= '0';
    when 6 => sel_f <= "00"; sel_e <= '0'; sel_a <= "00"; ce_e <= '0'; ce_a <= '1'; start_mult <= '0'; load <= '0'; done <= '0';

    when 7 => sel_f <= "01"; sel_e <= '0'; sel_a <= "00"; ce_e <= '1'; ce_a <= '0'; start_mult <= '0'; load <= '0'; done <= '0';
    when 8 => sel_f <= "01"; sel_e <= '0'; sel_a <= "00"; ce_e <= '0'; ce_a <= '0'; start_mult <= '1'; load <= '0'; done <= '0';
    when 9 => sel_f <= "01"; sel_e <= '0'; sel_a <= "00"; ce_e <= '0'; ce_a <= '0'; start_mult <= '0'; load <= '0'; done <= '0';
    when 10 => sel_f <= "01"; sel_e <= '0'; sel_a <= "00"; ce_e <= '0'; ce_a <= '1'; start_mult <= '0'; load <= '0'; done <= '0';

    when 11 => sel_f <= "10"; sel_e <= '0'; sel_a <= "00"; ce_e <= '1'; ce_a <= '0'; start_mult <= '0'; load <= '0'; done <= '0';
    when 12 => sel_f <= "10"; sel_e <= '0'; sel_a <= "00"; ce_e <= '0'; ce_a <= '0'; start_mult <= '1'; load <= '0'; done <= '0';
    when 13 => sel_f <= "10"; sel_e <= '0'; sel_a <= "00"; ce_e <= '0'; ce_a <= '0'; start_mult <= '0'; load <= '0'; done <= '0';
    when 14 => sel_f <= "10"; sel_e <= '0'; sel_a <= "00"; ce_e <= '0'; ce_a <= '1'; start_mult <= '0'; load <= '0'; done <= '0';

    when 15 => sel_f <= "11"; sel_e <= '0'; sel_a <= "00"; ce_e <= '1'; ce_a <= '0'; start_mult <= '0'; load <= '0'; done <= '0';
    when 16 => sel_f <= "11"; sel_e <= '0'; sel_a <= "00"; ce_e <= '0'; ce_a <= '0'; start_mult <= '1'; load <= '0'; done <= '0';
    when 17 => sel_f <= "11"; sel_e <= '0'; sel_a <= "00"; ce_e <= '0'; ce_a <= '0'; start_mult <= '0'; load <= '0'; done <= '0';
    when 18 => sel_f <= "11"; sel_e <= '0'; sel_a <= "00"; ce_e <= '0'; ce_a <= '1'; start_mult <= '0'; load <= '0'; done <= '0';

    when 19 => sel_f <= "00"; sel_e <= '0'; sel_a <= "01"; ce_e <= '1'; ce_a <= '0'; start_mult <= '0'; load <= '0'; done <= '0';
    when 20 => sel_f <= "00"; sel_e <= '0'; sel_a <= "01"; ce_e <= '0'; ce_a <= '0'; start_mult <= '1'; load <= '0'; done <= '0';
    when 21 => sel_f <= "00"; sel_e <= '0'; sel_a <= "01"; ce_e <= '0'; ce_a <= '0'; start_mult <= '0'; load <= '0'; done <= '0';
    when 22 => sel_f <= "00"; sel_e <= '0'; sel_a <= "01"; ce_e <= '0'; ce_a <= '1'; start_mult <= '0'; load <= '0'; done <= '0';

    when 23 => sel_f <= "00"; sel_e <= '1'; sel_a <= "10"; ce_e <= '1'; ce_a <= '0'; start_mult <= '0'; load <= '0'; done <= '0';
    when 24 => sel_f <= "00"; sel_e <= '0'; sel_a <= "10"; ce_e <= '0'; ce_a <= '0'; start_mult <= '1'; load <= '0'; done <= '0';
    when 25 => sel_f <= "00"; sel_e <= '0'; sel_a <= "10"; ce_e <= '0'; ce_a <= '0'; start_mult <= '0'; load <= '0'; done <= '0';
    when 26 => sel_f <= "00"; sel_e <= '0'; sel_a <= "10"; ce_e <= '0'; ce_a <= '1'; start_mult <= '0'; load <= '0'; done <= '0';
  end case;

  if reset = '1' then current_state <= 0;
  elsif clk'event and clk = '1' then
    case current_state is
      when 0 => if start = '0' then current_state <= 1; end if;
      when 1 => if start = '1' then current_state <= 2; end if;
      when 2 => current_state <= 3;

      when 3 => current_state <= 4;
      when 4 => current_state <= 5;
      when 5 => if mult_done = '1' then current_state <= 6; end if;
      when 6 => current_state <= 7;

      when 7 => current_state <= 8;
      when 8 => current_state <= 9;
      when 9 => if mult_done = '1' then current_state <= 10; end if;
      when 10 => current_state <= 11;

      when 11 => current_state <= 12;
      when 12 => current_state <= 13;
      when 13 => if mult_done = '1' then current_state <= 14; end if;
      when 14 => current_state <= 15;

      when 15 => current_state <= 16;
      when 16 => current_state <= 17;
      when 17 => if mult_done = '1' then current_state <= 18; end if;
      when 18 => current_state <= 19;

      when 19 => current_state <= 20;
      when 20 => current_state <= 21;
      when 21 => if mult_done = '1' then current_state <= 22; end if;
      when 22 => current_state <= 23;

      when 23 => current_state <= 24;
      when 24 => current_state <= 25;
      when 25 => if mult_done = '1' then current_state <= 26; end if;
      when 26 => current_state <= 0;
    end case;
  end if;
  end process control_unit;

end circuit;


 